/*
    *
    * This file is a part of CorePad.
    * A document editor for CuboCore Application Suite
    * Copyright 2019 CuboCore Group
    *

    *
    * This program is free software; you can redistribute it and/or modify
    * it under the terms of the GNU General Public License as published by
    * the Free Software Foundation; either version 3 of the License, or
    * (at your option) any later version.
    *

    *
    * This program is distributed in the hope that it will be useful,
    * but WITHOUT ANY WARRANTY; without even the implied warranty of
    * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    * GNU General Public License for more details.
    *

    *
    * You should have received a copy of the GNU General Public License
    * along with this program; if not, write to the Free Software
    * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
    * MA 02110-1301, USA.
    *
*/


#include <QFile>
#include <QFileDialog>
#include <QShortcut>
#include <QFileInfo>
#include <QDate>
#include <QMessageBox>
#include <QDebug>

#include <cprime/infofunc.h>
#include <cprime/filefunc.h>
#include <cprime/themefunc.h>
#include <cprime/shareit.h>
#include <cprime/pinit.h>

#include "corepad.h"
#include "ui_corepad.h"
#include "ceditor.h"
#include "ccolorscheme.h"


bool isBrightColor(QColor dst, QColor src)
{
    if (abs(dst.lightness() - src.lightness()) > 64) {
        return true;
    } else {
        return false;
    }
}

corepad::corepad(QWidget *parent) : QWidget(parent), ui(new Ui::corepad)
{
    ui->setupUi(this);

    tmpCount = -1;

    QPalette p(palette());
    p.setColor(QPalette::Base, Qt::transparent);
    setPalette(p);

    loadSettings();
    startSetup();
    shotcuts();
}

corepad::~corepad()
{
    delete ui;
}

/**
 * @brief Setup ui elements
 */
void corepad::startSetup()
{
    // Set stylesheet
    ui->sideView->setStyleSheet(CPrime::ThemeFunc::getSideViewStyleSheet());
    ui->sideView->setFixedWidth(sideViewSize);

    // Set window size
    if (touch) {
        this->setWindowState(Qt::WindowMaximized);
    } else {
        this->resize(800, 500);
    }

    ui->page->setCurrentIndex(0);

    // All toolbutton icon size
    for (QToolButton *b : this->findChildren<QToolButton *>()) {
        if (b) {
            b->setIconSize(sideViewIconSize);
        }
    }
    on_appTitle_clicked();
    connect(ui->cOpen, SIGNAL(clicked()), this, SLOT(openFileDialog()));
    connect(ui->addFiles, SIGNAL(clicked()), this, SLOT(openFileDialog()));

    ui->topBar->setVisible(false);
    ui->cSave->setEnabled(false);
    ui->cSaveAs->setEnabled(false);
    ui->search->setEnabled(false);
    ui->addDate->setEnabled(false);
    ui->cPaste->setEnabled(false);
    ui->pinIt->setEnabled(false);
    ui->shareIt->setEnabled(false);
}

/**
 * @brief Loads application settings
 */
void corepad::loadSettings()
{
    // Sideview is visible in default
    sideView = 1;

    // true = dark
    // false = light
    QString theme = "dark";
    QString schemeName = QString(":/syntax/%1.colorscheme").arg(theme);
    scheme = new CColorScheme(schemeName);

    QPalette pltt(palette());
    QColor mColor = scheme->getForegroundColor();

    if (!isBrightColor(pltt.color(QPalette::Window), mColor)) {
        if (pltt.color(QPalette::Window).lightness() < 128) {
            theme = "dark";
        } else {
            theme = "light";
        }
    }

    qDebug() << "Theme Name " << theme;

    schemeName = QString(":/syntax/%1.colorscheme").arg(theme);
    scheme = new CColorScheme(schemeName);

    sideViewSize = sm->value("CoreApps", "SideViewSize");
    sideViewIconSize = sm->value("CoreApps", "SideViewIconSize");
    touch = sm->value("CoreApps", "TouchMode");
    textViewFont = sm->value("CorePad", "Font");
}

void corepad::shotcuts()
{
    QShortcut *shortcut;
    shortcut = new QShortcut(QKeySequence(Qt::CTRL + Qt::Key_N), this);
    connect(shortcut, &QShortcut::activated, this, &corepad::on_cNew_clicked);
    shortcut = new QShortcut(QKeySequence(Qt::CTRL + Qt::Key_O), this);
    connect(shortcut, &QShortcut::activated, this, &corepad::openFileDialog);
    shortcut = new QShortcut(QKeySequence(Qt::CTRL + Qt::Key_S), this);
    connect(shortcut, &QShortcut::activated, this, &corepad::on_cSave_clicked);
    shortcut = new QShortcut(QKeySequence(Qt::CTRL + Qt::SHIFT + Qt::Key_S), this);
    connect(shortcut, &QShortcut::activated, this, &corepad::on_cSaveAs_clicked);
    shortcut = new QShortcut(QKeySequence(Qt::CTRL + Qt::Key_Z), this);
    connect(shortcut, &QShortcut::activated, this, &corepad::on_cUndo_clicked);
    shortcut = new QShortcut(QKeySequence(Qt::CTRL + Qt::Key_Y), this);
    connect(shortcut, &QShortcut::activated, this, &corepad::on_cRedo_clicked);
    shortcut = new QShortcut(QKeySequence(Qt::CTRL + Qt::Key_C), this);
    connect(shortcut, &QShortcut::activated, this, &corepad::on_cCopy_clicked);
    shortcut = new QShortcut(QKeySequence(Qt::CTRL + Qt::Key_X), this);
    connect(shortcut, &QShortcut::activated, this, &corepad::on_cCut_clicked);
    shortcut = new QShortcut(QKeySequence(Qt::CTRL + Qt::Key_V), this);
    connect(shortcut, &QShortcut::activated, this, &corepad::on_cPaste_clicked);
    shortcut = new QShortcut(QKeySequence(Qt::CTRL + Qt::Key_Q), this);
    connect(shortcut, &QShortcut::activated, this, &corepad::quitClicked);
    shortcut = new QShortcut(QKeySequence(Qt::CTRL + Qt::Key_F), this);
    connect(shortcut, &QShortcut::activated, this, &corepad::on_search_clicked);
}

bool corepad::initializeNewTab(const QString &filePath)
{
    if (!filePath.count()) {
        CPrime::InfoFunc::messageEngine("CorePad", "corepad", "Empty file path!!!", "Nothing to worry, select a valid file.", this);
        return false;
    }

    ui->page->setCurrentIndex(1);

    int tabsCount = ui->notes->count();

    if (tabsCount < 4) {
        CEditor *edit = nullptr;

        // Check whether the file is already opened
        for (int i = 0; i < ui->notes->count(); i++) {
            CEditor *tmp = static_cast<CEditor *>(ui->notes->widget(i));

            if (tmp->workFilePath() == filePath) {
                edit = tmp;
                break;
            }
        }

        if (!edit) {
            QFont font = textViewFont;
            edit = new CEditor(filePath, scheme, this);
            edit->setFocusPolicy(Qt::ClickFocus);
            edit->setFont(font);
            edit->lineNumberArea()->setFont(font);

            connect(edit, &CEditor::fileLoaded, this, &corepad::updateTab);
            connect(edit, &CEditor::unsavedChanges, this, &corepad::updateTab);
            ui->notes->addTab(edit, filePath);
        }

        edit->loadFile(filePath);
        edit->setFocus();
        ui->notes->setCurrentWidget(edit);
        setWindowTitle(ui->notes->tabText(ui->notes->currentIndex()) + " - CorePad");
        return true;
    } else {
        // Function from utilities.cpp
        CPrime::InfoFunc::messageEngine("CorePad", "corepad", "Reached page limit", " maximum tabs is 4. Close some tab.", this);
    }

    return false; //Return an exception - There is some other causes happening.
}

void corepad::sendFiles(const QStringList &paths)
{
    if (paths.count()) {
        for (QString str : paths) {
            if (CPrime::FileUtils::isFile(str)) {
                initializeNewTab(str);
            } else {
                qDebug() << "func(sendFiles) : Error : Invalid filepath got :" << str;
            }
        }
    }
}

void corepad::on_appTitle_clicked()
{
    if (sideView && touch) {
        // all button without text
        for (QToolButton *b : ui->sideView->findChildren<QToolButton *>()) {
            b->setToolButtonStyle(Qt::ToolButtonIconOnly);
            b->setFixedWidth(sideViewIconSize.width() + 6);
        }

        ui->sideView->setFixedWidth(sideViewIconSize.width() + 6);
        sideView = 0;
    } else {
        // all button with text
        for (QToolButton *b : ui->sideView->findChildren<QToolButton *>()) {
            b->setToolButtonStyle(Qt::ToolButtonTextBesideIcon);
            b->setFixedWidth(sideViewSize);
        }

        ui->sideView->setFixedWidth(sideViewSize);
        sideView = 1;
    }
}

void corepad::on_cNew_clicked()
{
    initializeNewTab(QString("untitled%1.txt").arg(++tmpCount));
}

bool corepad::on_cSave_clicked()
{
    CEditor *curr = currentEditor();

    if (curr == nullptr) {
        return false;
    }

    return curr->saveFile();
}

void corepad::on_cSaveAs_clicked()
{
    CEditor *curr = currentEditor();

    if (curr == nullptr) {
        return;
    }

    curr->saveFile(true);
}

void corepad::on_cCopy_clicked()
{
    currentEditor()->copy();
}

void corepad::on_cPaste_clicked()
{
    currentEditor()->paste();
}

void corepad::on_cCut_clicked()
{
    currentEditor()->cut();
}

void corepad::on_cUndo_clicked()
{
    currentEditor()->undo();
}

void corepad::on_cRedo_clicked()
{
    currentEditor()->redo();
}

void corepad::on_addDate_clicked()
{
    CEditor *text = currentEditor();
    text->insertPlainText(QDate::currentDate().toString("dd/MM/yyyy"));
}

void corepad::on_search_clicked()
{
    if (ui->topBar->isVisible()) {
        ui->topBar->setVisible(false);
        ui->searchHere->clear();
    } else {
        ui->topBar->setVisible(true);
        ui->searchHere->setFocus();
        CEditor *curr = currentEditor();
        QString search = curr->textCursor().selectedText();
        ui->searchHere->setText(search);
    }
}

void corepad::on_pinIt_clicked()
{
    QString workFilePath = currentEditor()->workFilePath();

    if (workFilePath.count()) {
        if (!QFileInfo(workFilePath).exists()) {
            // Function from utilities.cpp
            CPrime::InfoFunc::messageEngine("CorePad", "corepad", "File not exists or saved", "Please choose a valid file or save the file first.", this);
            return;
        }

        pinit pit;
        pit.callPinIt(this, workFilePath);
    }
}

void corepad::textCopyAvailable(bool b)
{
    ui->cCopy->setEnabled(b);
    ui->cCut->setEnabled(b);
}

void corepad::textUndoAvailable(bool b)
{
    ui->cUndo->setEnabled(b);
}

void corepad::textRedoAvailable(bool b)
{
    ui->cRedo->setEnabled(b);
}

void corepad::on_searchHere_textChanged(const QString &arg1)
{
    CEditor *curr = currentEditor();

    if (curr->textCursor().atEnd()) {
        curr->moveCursor(QTextCursor::Start);
    } else {
        curr->moveCursor(QTextCursor::Start, QTextCursor::KeepAnchor);
    }

    findS(arg1, false);
}

void corepad::on_nextW_clicked()
{
    findS(ui->searchHere->text(), false);
}

void corepad::on_previousW_clicked()
{
    findS(ui->searchHere->text(), true, QTextDocument::FindBackward);
}

void corepad::on_replaceT_clicked()
{
    CEditor *curr = currentEditor();

    if (curr->textCursor().selectedText() == ui->searchHere->text()) {
        curr->insertPlainText(ui->replaceWith->text());
    }

    on_nextW_clicked();
}

void corepad::on_replaceAllT_clicked()
{
    CEditor *curr = currentEditor();
    QString selected = curr->textCursor().selectedText();

    if (selected.count()) {
        if (selected == ui->searchHere->text()) {
            curr->insertPlainText(ui->replaceWith->text());
        }
    }

    bool found = curr->find(ui->searchHere->text());

    while (found) {
        curr->insertPlainText(ui->replaceWith->text());
        found = curr->find(ui->searchHere->text());
    }
}

void corepad::on_notes_currentChanged(int index)
{
    if (ui->notes->count() > 0) {
        CEditor *curr = currentEditor();

        if (!curr) {
            return;
        }

        textRedoAvailable(curr->isUndoRedoEnabled());
        textUndoAvailable(curr->isUndoRedoEnabled());
        on_searchHere_textChanged(ui->searchHere->text());

        ui->cSave->setEnabled(true);
        ui->cSaveAs->setEnabled(true);
        ui->search->setEnabled(true);
        ui->addDate->setEnabled(true);
        ui->cPaste->setEnabled(true);
        ui->pinIt->setEnabled(true);
        ui->shareIt->setEnabled(true);

        setWindowTitle(ui->notes->tabText(index) + " - CorePad");
    } else {
        ui->cSave->setEnabled(false);
        ui->cSaveAs->setEnabled(false);
        ui->search->setEnabled(false);
        ui->addDate->setEnabled(false);
        ui->cPaste->setEnabled(false);
        ui->pinIt->setEnabled(false);
        ui->shareIt->setEnabled(false);
        setWindowTitle("CorePad");
    }
}

void corepad::on_notes_tabCloseRequested(int index)
{
    ui->notes->setCurrentIndex(index);
    closeTab(index);
}

void corepad::updateTab(QString file)
{
    CEditor *curr = nullptr;
    int currIndex = -1;

    for (int i = 0; i < ui->notes->count(); i++) {
        CEditor *tmp = static_cast<CEditor *>(ui->notes->widget(i));

        if (tmp->workFilePath() == file) {
            curr = tmp;
            currIndex = i;
            break;
        }
    }

    if (!curr) {
        return;
    }

    bool changes = curr->hasChange();
    ui->notes->setTabText(currIndex, (changes ? "*" : "") + QFileInfo(file).fileName());
    setWindowTitle(ui->notes->tabText(currIndex) + " - CorePad");
}

void corepad::quitClicked()
{
    int tabCount = ui->notes->count();
    int cc = 0;

    for (int i = 0; i < tabCount; i++) {
        if (closeTab(i - cc)) {
            cc++;
        }
    }

    if (ui->notes->count() == 0) {
        if (this->close()) {
            deleteLater();
        }
    } else {
        CPrime::InfoFunc::messageEngine("CorePad", "corepad", "Can't close", "There are still some changes needs to be saved.", this);
    }
}

bool corepad::saveTo(const QString &filePath)
{
    QFile file(filePath);

    if (!file.open(QIODevice::Truncate | QIODevice::Text | QIODevice::WriteOnly)) {
        return false;
    } else {
        file.write(currentEditor()->toPlainText().toLocal8Bit());
        file.close();
        return true;
    }
}

void corepad::findS(QString searchS, bool reverse, QTextDocument::FindFlags flag)
{
    if (!searchS.isEmpty()) {
        CEditor *curr = currentEditor();
        QTextCursor cr;
        bool found = curr->find(searchS, flag);

        if (!found && !reverse) {
            cr = curr->textCursor();
            cr.movePosition(QTextCursor::Start);
            curr->setTextCursor(cr);
            curr->find(searchS, flag);
        }

        if (reverse) {
            cr.movePosition(QTextCursor::End);

            if (!(cr.position() < 0)) {
                curr->setTextCursor(cr);
            }
        }

    }
}

QStringList corepad::unsavedFiles()
{
    QStringList unsaved;

    for (int i = 0; i < ui->notes->count(); i++) {
        CEditor *tmp = static_cast<CEditor *>(ui->notes->widget(i));

        if (tmp->hasChange()) {
            unsaved << tmp->workFilePath();
        }
    }

    return unsaved;
}

bool corepad::closeTab(int index)
{
    CEditor *edit = static_cast<CEditor *>(ui->notes->widget(index));

    if (edit == nullptr) {
        return false;
    }

    QString filePath = edit->workFilePath();

    if (edit->hasChange()) {
        int reply = QMessageBox::question(this, tr("Unsaved Changes?"),
                                          QString(tr("This file has unsaved changes.\nDo you want to discard it?\n\n%1"))
                                          .arg(filePath), QMessageBox::Yes | QMessageBox::No, QMessageBox::No);

        if (reply != QMessageBox::Yes) {
            return false;
        }
    }

    if (QFileInfo(filePath).exists()) {
        qDebug() << filePath << "Saved to recent " << CPrime::InfoFunc::saveToRecent("corepad", filePath);
    }

    edit->deleteLater();
    ui->notes->removeTab(index);
    on_notes_currentChanged(index);
    return true;
}

CEditor *corepad::currentEditor()
{
    if (ui->notes->count() < 1) {
        return nullptr;
    }

    return static_cast<CEditor *>(ui->notes->currentWidget());
}

void corepad::closeEvent(QCloseEvent *event)
{
    QStringList unsaved = unsavedFiles();

    if (unsaved.isEmpty()) {
        while (ui->notes->count()) {
            closeTab(ui->notes->count() - 1);
        }

        QWidget::closeEvent(event);
        return;
    }

    // Otherwise, ask the user what to do.
    QMessageBox::StandardButton but = QMessageBox::question(
                                          this,
                                          tr("Save Changes before closing?"),
                                          QString(tr("There are unsaved changes.\nDo you want save them before you close the editor?\n\n%1")).arg(unsaved.join("\n")),
                                          QMessageBox::Yes | QMessageBox::No | QMessageBox::Cancel,
                                          QMessageBox::No);

    if (but == QMessageBox::Cancel) {
        event->ignore();
        return;
    } else if (but == QMessageBox::Yes) {
        // Save all opened files
        bool ok = true;

        for (int i = 0; i < ui->notes->count() && ok; i++) {
            CEditor *tmp = static_cast<CEditor *>(ui->notes->widget(i));

            if (tmp->hasChange()) {
                ok = ok && tmp->saveFile();
            }
        }

        if (!ok) {
            //cancelled by user
            event->ignore();
            return;
        }
    }

    QWidget::closeEvent(event);
}

void corepad::on_shareIt_clicked()
{
    QString workFilePath = currentEditor()->workFilePath();

    if (workFilePath.count()) {
        if (!QFileInfo(workFilePath).exists()) {
            // Function from utilities.cpp
            CPrime::InfoFunc::messageEngine("CorePad", "corepad", "File not exists or saved", "Please choose a valid file or save the file first.", this);
            return;
        }

        shareit::shareFiles( this, QStringList() << workFilePath );
    }
}

void corepad::openFileDialog()
{
    QStringList fileP = QFileDialog::getOpenFileNames(this, tr("Open File"), QDir::homePath() + "/Documents", tr("All Files (*)"));

    if (!fileP.count()) {
        // Function from utilities.cpp
        CPrime::InfoFunc::messageEngine("CorePad", "corepad", "Empty File Name", "Select a valid file.", this);
        return;
    } else {
        int failed = 0;
        for (QString path : fileP) {
            bool open = initializeNewTab(path);

            if (!open) {
                failed++;
            }
        }

        if (failed) {
            CPrime::InfoFunc::messageEngine("CorePad", "corepad", "Can't open some file(s)", "Invalid file or corrupted.", this);
        } else {
            CPrime::InfoFunc::messageEngine("CorePad", "corepad", "File(s) Opened successfully", "Do your changes", this);
        }
    }
}

void corepad::on_newTextW_clicked()
{
    initializeNewTab(QString("untitled%1.txt").arg(++tmpCount));
}
